unit nvMultiBitmapHDS;
{ ============================================================================= }
{                               UNIT nvMultiBitmapHDS                           }
{ Description : Provides the ability to generate a terrain using muliple        }
{ bitmaps. This can be useful if you for example wanted a single terrain with   }
{ some parts of the terrian having greater height detail than others (eg        }
{ could have a large less detailed greyscale bitmap for a land mass but for     }
{ town have a less scaled, more detailed bitmap overlapping the larger one.     }
{                                                                               }
{ Created May 2004 Skinhat                                                      }
{                                                                               }
{ ============================================================================= }

interface
uses Windows, SysUtils, Classes, Graphics, VectorGeometry, GLCrossPlatform,GLHeightData;
type

  { Collection of height datasource (greyscale bitmaps).}
  TElevationMap = class(TCollectionItem)
  private
    FScanLineCache : array of PByteArray;
    FWidth:integer;
    FPicture : TPicture;
    FBitmap:TBitmap;
    procedure setWidth(width:integer);
  protected
    procedure SetPicture(const val : TPicture);
    procedure OnPictureChanged(Sender : TObject);
    procedure CreateMonochromeBitmap(size : Integer);
    procedure FreeMonochromeBitmap;

  public
    x,y:integer;
    scale:single;
    constructor create(Collection:TCollection);  override;
    destructor destroy; override;
    function doesoverlap(inx,iny,inwidth:integer):Boolean;
    function getScanLine(iny:integer):PByteArray;
    property width:integer read FWidth write SetWidth;
   {: The picture serving as Height field data reference.<p>
      The picture is (if not already) internally converted to a 8 bit
      bitmap (grayscale). For better performance and to save memory,
      feed it this format! }
    property Picture : TPicture read FPicture write SetPicture;

  end;

  TElevationMaps = class(TCollection)
    function GetItem(Index: Integer): TElevationMap;
    procedure SetItem(Index: Integer; const Value: TElevationMap);
  public
    constructor create;
    function Add : TElevationMap;
    property Items[ Index : Integer ] : TElevationMap read GetItem write SetItem; default;
    function doesOverlap(inx,iny,inwidth:integer):Boolean;
    function width:integer;
  end;


	// TGLMultiBitmapHDS
	//

   {: Bitmap-based Height Data Source.<p>
      The image is automatically wrapped if requested data is out of picture size,
      or if requested data is larger than the picture.<p>
      The internal format is an 8 bit bitmap whose dimensions are a power of two,
      if the original image does not comply, it is StretchDraw'ed on a monochrome
      (gray) bitmap. }
  TGLMultiBitmapHDS = class (THeightDataSource)
  private
    { Private Declarations }
    //FScanLineCache : array of PByteArray;
   // FBitmap : Graphics.TBitmap;
    //FPicture : TPicture;
  //  FScale: Single;
    FElevationMaps:TElevationMaps ;
  protected
    { Protected Declarations }
    //procedure SetPicture(const val : TPicture);
    procedure OnPictureChanged(Sender : TObject);

  //  procedure CreateMonochromeBitmap(size : Integer);
  //  procedure FreeMonochromeBitmap;
   // function GetScanLine(y : Integer) : PByteArray;

    procedure StartPreparingData(heightData : THeightData); override;

  public
    { Public Declarations }
    constructor Create(ElevationMaps:TElevationMaps);
    destructor Destroy; override;
    procedure MarkDirty(const area : TGLRect); override;

  published
        { Published Declarations }
   {: The picture serving as Height field data reference.<p>
      The picture is (if not already) internally converted to a 8 bit
      bitmap (grayscale). For better performance and to save memory,
      feed it this format! }
  //  property Picture : TPicture read FPicture write SetPicture;
   {: If true the height field is wrapped indefinetely. }

   // property MaxPoolSize;
   // property Scale:Single read FScale write FScale;

  end;

implementation

uses
  GLMisc,ApplicationFileIO, GLUtils;

constructor TElevationMap.create(Collection:TCollection);
begin
  x:=0;
  y:=0;
  scale:=1;
  Fwidth:=-1;
  FPicture:=TPicture.create;
  FPicture.OnChange:=OnPictureChanged;
  inherited create(Collection);
end;

destructor TElevationMap.destroy;
begin
  FPicture.Free;
end;

function TElevationMap.getScanLine(iny:integer):PByteArray;
var
  iScan:integer;
begin
  { Is iny in range }
  if (iny>=y) and (iny<y+width) then
  begin
    iScan:=trunc((iny-y)/Scale);
    Result:=FScanLineCache[iScan];
    if not Assigned(Result) then
    begin
      Result:=FBitmap.ScanLine[iScan];
      FScanLineCache[iScan]:=Result;
    end;
  end
   // Result:=FBitmap.scanline[trunc((iny-y)/Scale)]
  else
    Result:=nil;
end;

function TElevationMap.doesoverlap(inx,iny,inwidth:integer):Boolean;
var
  R1,R2:TRect;
  outRect:TRect;
begin
  R1.Left:=x;
  R1.Top:=y;
  R1.Bottom:=trunc(y+width*scale);
  R1.Right:=trunc(x+width*scale);

  R2.Left:=inx;
  R2.Top:=iny;
  R2.Bottom:=iny+inwidth;
  R2.Right:=inx+inwidth;
  Result:=IntersectRect(outRect,R1, R2);


  { Check if any edge is within the map }
 // Result:=  ((inx>=x) and (inx<=x+width) and (iny<y+width) and (iny>y)) or
 //           ((inx+width>=x) and (iny<y+width))
end;

procedure TElevationMap.setWidth(width:integer);
begin
  FWidth:=width;
  if assigned(Fbitmap) then
    Scale:=width/picture.bitmap.width
  else
    Scale:=-1;

end;

procedure TElevationMap.SetPicture(const val : TPicture);
begin
   FPicture.Assign(val);
end;

(*procedure TElevationMap.loadfromfile(filename:String);
begin
  if not assigned(bitmap) then
    bitmap:=TBitmap.create;
  bitmap.loadfromfile(filename);
  if width=-1 then
    width:=bitmap.Width;
end;     *)

procedure TElevationMap.OnPictureChanged(Sender : TObject);
var
  oldPoolSize, size : Integer;
begin
   // cleanup pool
   {!!!!!! dont forget about this }

   
 (* oldPoolSize:=MaxPoolSize;
  MaxPoolSize:=0;
  CleanUp;
  MaxPoolSize:=oldPoolSize; *)
   // prepare MonoChromeBitmap
  FreeMonochromeBitmap;
  size:=Picture.Width;
  if size>0 then
    CreateMonochromeBitmap(size);
  if width<=0 then
    width:=Picture.Width;
end;

procedure TElevationMap.CreateMonochromeBitmap(size : Integer);
type
   TPaletteEntryArray = array [0..255] of TPaletteEntry;
   PPaletteEntryArray = ^TPaletteEntryArray;
   TLogPal = record
      lpal : TLogPalette;
      pe : TPaletteEntryArray;
   end;

var
   x : Integer;
   logpal : TLogPal;
   hPal : HPalette;
begin
   //size:=RoundUpToPowerOf2(size);
   FBitmap:=Graphics.TBitmap.Create;
   FBitmap.PixelFormat:=pf8bit;
   FBitmap.Width:=size;
   FBitmap.Height:=size;
   for x:=0 to 255 do with PPaletteEntryArray(@logPal.lpal.palPalEntry[0])[x] do begin
      peRed:=x;
      peGreen:=x;
      peBlue:=x;
      peFlags:=0;
   end;
   with logpal.lpal do begin
      palVersion:=$300;
      palNumEntries:=256;
   end;
   hPal:=CreatePalette(logPal.lpal);
   Assert(hPal<>0);
   FBitmap.Palette:=hPal;
   // some picture formats trigger a "change" when drawed
   Picture.OnChange:=nil;
   try

      FBitmap.Canvas.StretchDraw(Rect(0, 0, size, size), Picture.Graphic);
   finally
      Picture.OnChange:=OnPictureChanged;
   end;
   SetLength(FScanLineCache, 0); // clear the cache
   SetLength(FScanLineCache, size);
end;

// FreeMonochromeBitmap
//
procedure TElevationMap.FreeMonochromeBitmap;
begin
   SetLength(FScanLineCache, 0);
   FBitmap.Free;
   FBitmap:=nil;
end;

constructor TElevationMaps.create;
begin
  inherited create(TElevationMap);
end;

function TElevationMaps.GetItem(Index: Integer): TElevationMap;
begin
  Result := TElevationMap(inherited GetItem(Index));
end;

procedure TElevationMaps.SetItem(Index: Integer; const Value: TElevationMap );
begin
  inherited SetItem(Index, Value);
end;

function TElevationMaps.Add : TElevationMap;
begin
  Result := TElevationMap(inherited Add);
end;

function TElevationMaps.doesOverlap(inx,iny,inwidth:integer):Boolean;
var
  i:integer;
begin
  i:=0;
  while (i<count) and ( not items[i].doesOverlap(inx,iny,inwidth)) do
    inc(i);
  result:=i<count;
end;

function TElevationMaps.width:integer;
var
  i:integer;
  elevationMap:TElevationMap;
begin
  result:=0;
  for i:=0 to count-1 do
  begin
    elevationMap:=items[i];
    if elevationMap.x+elevationMap.width>result then
      result:=elevationMap.x+elevationMap.width;
  end;
end;


// ------------------
// ------------------ TGLMultiBitmapHDS ------------------
// ------------------

// Create
//
constructor TGLMultiBitmapHDS.Create(ElevationMaps:TElevationMaps);
begin
  inherited Create(nil);
 // FPicture:=TPicture.Create;
//  FPicture.OnChange:=OnPictureChanged;
 // FScale:=1;
  FElevationMaps:=ElevationMaps;
end;

// Destroy
//
destructor TGLMultiBitmapHDS.Destroy;
begin
  inherited Destroy;
  //FreeMonochromeBitmap;
 // FPicture.Free;
end;

// SetPicture
//
//procedure TGLMultiBitmapHDS.SetPicture(const val : TPicture);
//begin
   //FPicture.Assign(val);
//end;

// OnPictureChanged
//
procedure TGLMultiBitmapHDS.OnPictureChanged(Sender : TObject);
var
   oldPoolSize, size : Integer;
begin
   // cleanup pool
   oldPoolSize:=MaxPoolSize;
   MaxPoolSize:=0;
   CleanUp;
   MaxPoolSize:=oldPoolSize;
   // prepare MonoChromeBitmap
  // FreeMonochromeBitmap;
  // size:=Picture.Width;
  // if size>0 then
   //   CreateMonochromeBitmap(size);
end;



// MarkDirty
//
procedure TGLMultiBitmapHDS.MarkDirty(const area : TGLRect);
begin
   inherited;
  // FreeMonochromeBitmap;
  // if Picture.Width>0 then
  //    CreateMonochromeBitmap(Picture.Width);
end;

// CreateMonochromeBitmap
//
(*procedure TGLMultiBitmapHDS.CreateMonochromeBitmap(size : Integer);
type
   TPaletteEntryArray = array [0..255] of TPaletteEntry;
   PPaletteEntryArray = ^TPaletteEntryArray;
   TLogPal = record
      lpal : TLogPalette;
      pe : TPaletteEntryArray;
   end;

var
   x : Integer;
   logpal : TLogPal;
   hPal : HPalette;
begin
   size:=RoundUpToPowerOf2(size);
   FBitmap:=Graphics.TBitmap.Create;
   FBitmap.PixelFormat:=pf8bit;
   FBitmap.Width:=size;
   FBitmap.Height:=size;
   for x:=0 to 255 do with PPaletteEntryArray(@logPal.lpal.palPalEntry[0])[x] do begin
      peRed:=x;
      peGreen:=x;
      peBlue:=x;
      peFlags:=0;
   end;
   with logpal.lpal do begin
      palVersion:=$300;
      palNumEntries:=256;
   end;
   hPal:=CreatePalette(logPal.lpal);
   Assert(hPal<>0);
   FBitmap.Palette:=hPal;
   // some picture formats trigger a "change" when drawed
   Picture.OnChange:=nil;
   try
      FBitmap.Canvas.StretchDraw(Rect(0, 0, size, size), Picture.Graphic);
   finally
      Picture.OnChange:=OnPictureChanged;
   end;
   SetLength(FScanLineCache, 0); // clear the cache
   SetLength(FScanLineCache, size);
end;

// FreeMonochromeBitmap
//
procedure TGLMultiBitmapHDS.FreeMonochromeBitmap;
begin
   SetLength(FScanLineCache, 0);
   FBitmap.Free;
   FBitmap:=nil;
end;
     *)
// GetScanLine
//
{function TGLMultiBitmapHDS.GetScanLine(y : Integer) : PByteArray;
begin
   //y:=trunc(y / FScale);
   Result:=FScanLineCache[y];
   if not Assigned(Result) then begin
      Result:=FBitmap.ScanLine[y];
      FScanLineCache[y]:=Result;
   end;
end;   }

// StartPreparingData
//
procedure TGLMultiBitmapHDS.StartPreparingData(heightData : THeightData);
var
  i:integer;
  y, x : Integer;
  bmpSize, wrapMask : Integer;
  bitmapLine, rasterLine : PByteArray;
  oldType : THeightDataType;
  b : Byte;
  ElevationMap:TElevationMap;
  iScan:integer;
begin

  { Check to see if heightData is in bounds }
  { Go from back to front and check if overlap }
  if assigned(FElevationMaps) and FElevationMaps.doesOverlap(HeightData.XLeft, HeightData.YTop, HeightData.Size-1) then
  begin
    heightData.DataState:=hdsPreparing;
    HeightData.Allocate(hdtByte);
    for i:=FElevationMaps.count-1 downto 0 do
    begin
      ElevationMap:=FElevationMaps[i];
      { Does this map overlap the current height data }
      if ElevationMap.doesoverlap(HeightData.XLeft, HeightData.YTop, HeightData.Size-1) then
      begin
        for y:=HeightData.YTop to HeightData.YTop+HeightData.Size-1 do
        begin
          rasterLine:=HeightData.ByteRaster[y-HeightData.YTop];
          bitmapline:=ElevationMap.getScanLine(y);
          if assigned(bitmapLine) then
          begin
             // *BIG CAUTION HERE* : Don't remove the intermediate variable here!!!
             // or Delphi compiler will "optimize" to 32 bits access with clamping
             // resulting in possible reads of stuff beyon bitmapLine length!!!!
             for x:=HeightData.XLeft to HeightData.XLeft+HeightData.Size-1 do
             begin
               iScan:=trunc((x-ElevationMap.x)/ElevationMap.scale);
               if (iScan>=0) and (iScan<ElevationMap.Fbitmap.width) then
                 b:=bitmapLine[ElevationMap.Fbitmap.width-iScan-1]
               else
                 b:=0;
               rasterLine[x-HeightData.XLeft]:=b;
             end;
           end
           else
           begin
             for x:=HeightData.XLeft to HeightData.XLeft+HeightData.Size-1 do
             begin
               b:=0;
               rasterLine[x-HeightData.XLeft]:=b;
             end
           end;
        end;

      end;
    end;
    inherited;
  end
  else
    heightData.DataState:=hdsNone;

end;



end.
