unit uScaledBitmapHDS;
{ ============================================================================= }
{                               UNIT uScaledBitmapHDS                           }
{ Description: Allows the scaling of a greyscale height data source image by a  }
{ factor. This can be useful if you want trick Glscene into giving you          }
{ a tilesize of 1 instead of 8 ie just make the scale 8!                        }
{                                                                               }
{ Created May 2004 Skinhat                                                      }
{ ============================================================================= }
interface
uses Windows, SysUtils, Classes, Graphics, VectorGeometry, GLCrossPlatform,GLHeightData;
type

	// TGLBitmapHDS
	//
   {: Bitmap-based Height Data Source.<p>
      The image is automatically wrapped if requested data is out of picture size,
      or if requested data is larger than the picture.<p>
      The internal format is an 8 bit bitmap whose dimensions are a power of two,
      if the original image does not comply, it is StretchDraw'ed on a monochrome
      (gray) bitmap. }
	TGLScaledBitmapHDS = class (THeightDataSource)
	   private
	      { Private Declarations }
         FScanLineCache : array of PByteArray;
         FBitmap : Graphics.TBitmap;
         FPicture : TPicture;
         FInfiniteWrap : Boolean;
         FScale: Single;

	   protected
	      { Protected Declarations }
         procedure SetPicture(const val : TPicture);
         procedure OnPictureChanged(Sender : TObject);
         procedure SetInfiniteWrap(val : Boolean);

         procedure CreateMonochromeBitmap(size : Integer);
         procedure FreeMonochromeBitmap;
         function GetScanLine(y : Integer) : PByteArray;

         procedure StartPreparingData(heightData : THeightData); override;

	   public
	      { Public Declarations }
	      constructor Create(AOwner: TComponent); override;
         destructor Destroy; override;

         procedure MarkDirty(const area : TGLRect); override;

	   published
	      { Published Declarations }
         {: The picture serving as Height field data reference.<p>
            The picture is (if not already) internally converted to a 8 bit
            bitmap (grayscale). For better performance and to save memory,
            feed it this format! }
         property Picture : TPicture read FPicture write SetPicture;
         {: If true the height field is wrapped indefinetely. }
         property InfiniteWrap : Boolean read FInfiniteWrap write SetInfiniteWrap default True;

         property MaxPoolSize;
         property Scale:Single read FScale write FScale;

	end;
implementation
uses   GLMisc,ApplicationFileIO, GLUtils;

// ------------------
// ------------------ TGLScaledBitmapHDS ------------------
// ------------------

// Create
//
constructor TGLScaledBitmapHDS.Create(AOwner: TComponent);
begin
	inherited Create(AOwner);
   FPicture:=TPicture.Create;
   FPicture.OnChange:=OnPictureChanged;
   FInfiniteWrap:=True;
   FScale:=1;
end;

// Destroy
//
destructor TGLScaledBitmapHDS.Destroy;
begin
	inherited Destroy;
   FreeMonochromeBitmap;
   FPicture.Free;
end;

// SetPicture
//
procedure TGLScaledBitmapHDS.SetPicture(const val : TPicture);
begin
   FPicture.Assign(val);
end;

// OnPictureChanged
//
procedure TGLScaledBitmapHDS.OnPictureChanged(Sender : TObject);
var
   oldPoolSize, size : Integer;
begin
   // cleanup pool
   oldPoolSize:=MaxPoolSize;
   MaxPoolSize:=0;
   CleanUp;
   MaxPoolSize:=oldPoolSize;
   // prepare MonoChromeBitmap
   FreeMonochromeBitmap;
   size:=Picture.Width;
   if size>0 then
      CreateMonochromeBitmap(size);
end;

// SetInfiniteWrap
//
procedure TGLScaledBitmapHDS.SetInfiniteWrap(val : Boolean);
begin
   if FInfiniteWrap<>val then begin
      FInfiniteWrap:=val;
{$ifdef GLS_DELPHI_4}
      inherited MarkDirty;
{$else}
      MarkDirty;
{$endif}
   end;
end;

// MarkDirty
//
procedure TGLScaledBitmapHDS.MarkDirty(const area : TGLRect);
begin
   inherited;
   FreeMonochromeBitmap;
   if Picture.Width>0 then
      CreateMonochromeBitmap(Picture.Width);
end;

// CreateMonochromeBitmap
//
procedure TGLScaledBitmapHDS.CreateMonochromeBitmap(size : Integer);
type
   TPaletteEntryArray = array [0..255] of TPaletteEntry;
   PPaletteEntryArray = ^TPaletteEntryArray;
   TLogPal = record
      lpal : TLogPalette;
      pe : TPaletteEntryArray;
   end;

var
   x : Integer;
   logpal : TLogPal;
   hPal : HPalette;
begin
   size:=RoundUpToPowerOf2(size);
   FBitmap:=Graphics.TBitmap.Create;
   FBitmap.PixelFormat:=pf8bit;
   FBitmap.Width:=size;
   FBitmap.Height:=size;
   for x:=0 to 255 do with PPaletteEntryArray(@logPal.lpal.palPalEntry[0])[x] do begin
      peRed:=x;
      peGreen:=x;
      peBlue:=x;
      peFlags:=0;
   end;
   with logpal.lpal do begin
      palVersion:=$300;
      palNumEntries:=256;
   end;
   hPal:=CreatePalette(logPal.lpal);
   Assert(hPal<>0);
   FBitmap.Palette:=hPal;
   // some picture formats trigger a "change" when drawed
   Picture.OnChange:=nil;
   try
      FBitmap.Canvas.StretchDraw(Rect(0, 0, size, size), Picture.Graphic);
   finally
      Picture.OnChange:=OnPictureChanged;
   end;
   SetLength(FScanLineCache, 0); // clear the cache
   SetLength(FScanLineCache, size);
end;

// FreeMonochromeBitmap
//
procedure TGLScaledBitmapHDS.FreeMonochromeBitmap;
begin
   SetLength(FScanLineCache, 0);
   FBitmap.Free;
   FBitmap:=nil;
end;

// GetScanLine
//
function TGLScaledBitmapHDS.GetScanLine(y : Integer) : PByteArray;
begin
   y:=trunc(y / FScale);
   Result:=FScanLineCache[y];
   if not Assigned(Result) then begin
      Result:=FBitmap.ScanLine[y];
      FScanLineCache[y]:=Result;
   end;
end;

// StartPreparingData
//
procedure TGLScaledBitmapHDS.StartPreparingData(heightData : THeightData);
var
   y, x : Integer;
   bmpSize, wrapMask : Integer;
   bitmapLine, rasterLine : PByteArray;
   oldType : THeightDataType;
   b : Byte;
begin
   if FBitmap=nil then Exit;
   heightData.DataState:=hdsPreparing;
   bmpSize:=round(FBitmap.Width*FScale);
   wrapMask:=bmpSize-1;
   // retrieve data
   with heightData do begin
      if (not InfiniteWrap)
         and ((XLeft>=bmpSize) or (XLeft<0)
              or (YTop>=bmpSize) or (YTop<0)) then begin
         heightData.DataState:=hdsNone;
         Exit;
      end;
      oldType:=DataType;
      Allocate(hdtByte);
      for y:=YTop to YTop+Size-1 do begin
         bitmapLine:=GetScanLine(y and wrapMask);
         rasterLine:=ByteRaster[y-YTop];
         // *BIG CAUTION HERE* : Don't remove the intermediate variable here!!!
         // or Delphi compiler will "optimize" to 32 bits access with clamping
         // resulting in possible reads of stuff beyon bitmapLine length!!!!
         for x:=XLeft to XLeft+Size-1 do begin
            b:=bitmapLine[trunc(x/ FScale) and wrapMask];
            rasterLine[x-XLeft]:=b;
         end;
      end;
      if (oldType<>hdtByte) and (oldType<>hdtDefault) then
         DataType:=oldType;
   end;
   inherited;
end;



end.
